# EnSimpleStaging  |  Copyright(C), 2004, Enfold Systems, LLC
# see LICENSE.txt for details

import os, sys
from sets import Set

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing.ZopeTestCase.functional import Functional
from Products.PloneTestCase import PloneTestCase

for product in ('Archetypes',
                'PortalTransforms',
                'ZopeVersionControl',
                'CMFStaging',
                'EnSimpleStaging',
                'MimetypesRegistry',
                'MenuGenerator',
#                'SecureMailHost',
                ):
    PloneTestCase.installProduct(product)
del product

from cStringIO import StringIO
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.tests.utils import makeContent
from Products.EnSimpleStaging.Extensions.Install import setup_local_tools

PloneTestCase.setupPloneSite()

class TestZopeVersionControl(PloneTestCase.ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.app.manage_addProduct['OFSP'].manage_addFolder('afolder1')
        self.afolder1 = self.app.afolder1
        self.app.manage_addProduct['ZopeVersionControl'].addRepository('repo')
        self.repo = self.app.repo

    def testNonVersionable(self):
        self.afolder1.manage_addFolder('afolder11')
        self.afolder1.afolder11.__non_versionable__ = True
        self.afolder1.manage_addFolder('afolder12')
        self.repo.applyVersionControl(self.afolder1)
        vi = self.app.repo.getVersionInfo(self.afolder1)
        versobj = self.app.repo.getVersionOfResource(vi.history_id,
                                                 vi.version_id)
        result = versobj.restrictedTraverse('afolder11', None)
        self.failUnless(result, "did not find afolder11")
        result = versobj.restrictedTraverse('afolder12', None)
        self.failIf(result, "found misplaced %s")

##    def test_isResourceChanged(self):
##        self.repo.applyVersionControl(self.afolder1)
##        self.failIf(self.repo.isResourceChanged(self.afolder1))
##        self.afolder1.manage_changeProperties(title='new title')
##        get_transaction().commit(1)
##        # let's sleep a little
##        import time
##        time.sleep(1)
##        self.failUnless(self.repo.isResourceChanged(self.afolder1))
        

class TestStaging(PloneTestCase.PloneTestCase):

    P_COUNT = 13

    #  \index_html
    #  \doc1
    #  \folder1
    #     \folder11
    #     \folder12
    #     \doc11
    #  \folder2
    #     \folder21
    #     \doc21
    #     \index_html
    #     \folder22
    #        \folder221
    #           \doc2211

    _defaultStructure = [dict(id='index_html', type='Document'),
                         dict(id='doc1', type='Document'),
                         dict(id='folder1', type='Folder',
                              objs=[dict(id='doc11', type='Document'),
                                    dict(id='folder11', type='Folder'),
                                    dict(id='folder12', type='Folder'),
                                    ]),
                         dict(id='folder2', type='Folder',
                              objs=[dict(id='index_html', type='Document'),
                                    dict(id='doc21', type='Document'),
                                    dict(id='folder21', type='Folder'),
                                    dict(id='folder22', type='Folder',
                                         objs=[dict(id='folder221',
                                                    type='Folder',
                                                    objs=[dict(id='doc2211',
                                                               type='Document')]),
                                               ]),
                                    ]),
                         ]

    def _populateFolder(self, folder, structure=_defaultStructure):
        for item in structure:
            obj = makeContent(folder, item['type'], id=item['id'])
            self._populateFolder(obj, structure=item.get('objs', []))

    def _checkStructure(self, folder, structure=_defaultStructure,
                        path='', errors=None):
        if errors is None:
            errors = {}
        for item in structure:
            if item['id'] not in folder.objectIds():
                errors.setdefault(path, []).append((item['id'], 'NotFound'))
                continue
            obj = folder[item['id']]
            if obj.portal_type != item['type']:
                errors.setdefault(path, []).append((item, obj.portal_type,
                                                    'BadType'))
            self._checkStructure(obj, structure=item.get('objs', []),
                                 path=path + '/' + item['id'], errors=errors)
        return errors

    def afterSetUp(self):
        self.loginAsPortalOwner()
        self.qi = self.portal.portal_quickinstaller
        self.qi.installProduct('EnSimpleStaging')
        self.mt = self.portal.portal_membership
        self.wt = self.portal.portal_workspaces
        self.vt = self.portal.portal_versions
        self.acl_users = self.portal.acl_users
        self.repo = self.portal.VersionRepository
        self.home = self.mt.getHomeFolder()
        self.public = makeContent(self.portal, 'Folder', id='public_website')
        self.setupSkin()

    def setupSkin(self):
        self.portal._v_skindata = None
        self.portal.setupCurrentSkin()

    def subcommit(self):
        get_transaction().commit(1)
        # subcommiting causes the portal to be ghosted
        # we need to re-setup the skin
        self.setupSkin()

    def setupStagingAreas(self):
        for area in 'mkt adv sales'.split():
            areaTitle = area + ' staging area'
            self.home.invokeFactory('StagingArea', area, title=areaTitle)

    def setupNestedSite(self):
        setup_local_tools(self.portal, StringIO())

class TestEnSimpleStaging(TestStaging):

    def testSetup(self):
        self.setupStagingAreas()
        self.failUnless(hasattr(self.home, 'mkt'))
        stage = self.home.mkt
        self.failUnless(hasattr(stage, 'public'), 'Staging area has no public node')
        self.failUnless(hasattr(stage, 'private'), 'Staging area has no private node')
        self.failUnless(stage.public.portal_type == 'Sandbox' == stage.private.portal_type)

    def test_getStageItems(self):
        wt = self.wt
        self.failIf(wt.getStageItems())
        self.setupStagingAreas()
        numstages = len(wt.getStageItems())
        self.assertEquals(numstages, 3,
                          "Destinations unset but %s stages" % numstages)
        mkt = self.home.mkt
        adv = self.home.adv
        sales = self.home.sales
        mktID = mkt.UID()
        advID = adv.UID()
        salesID = sales.UID()
        public = self.public
        public.invokeFactory('Folder', id='marketing')
        public.invokeFactory('Folder', id='sales')
        public.marketing.invokeFactory('Folder', id='advertising')
        mkt.setStagePath('public_website/marketing')
        adv.setStagePath('public_website/marketing/advertising')
        sales.setStagePath('public_website/sales')
        tool_stages = wt.getStageItems()
        self.failUnless(tool_stages)
        # let's make our own list from the stages (and make sure it's a list)
        tool_stages = list(tool_stages)
        tool_stages.sort()
        expect = [('dst_%s' % mktID, 'mkt staging area (destination)', 'public_website/marketing'),
                  ('dst_%s' % salesID, 'sales staging area (destination)', 'public_website/sales'),
                  ('dst_%s' % advID, 'adv staging area (destination)', 'public_website/marketing/advertising'),
                  ('src_%s' % mktID, 'mkt staging area (source)', 'Members/portal_owner/mkt/public'),
                  ('src_%s' % advID, 'adv staging area (source)', 'Members/portal_owner/adv/public'),
                  ('src_%s' % salesID, 'sales staging area (source)', 'Members/portal_owner/sales/public')]
        expect.sort()
        # I could have done a direct comparisson of the 2 lists, but
        # this way it's easier to spot the differences
        for i1, i2 in zip(expect, tool_stages):
            self.assertEquals(i1, i2)

    def testStagePublishing(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        adv = self.home.adv
        public = self.public
        public.invokeFactory('Folder', id='marketing')
        public.invokeFactory('Folder', id='advertising')
        # Create index_html on public_website
        # to see if we trigger a bug in staging.
        public.invokeFactory('Document', id='index_html')

        mkt.setStagePath('public_website/marketing')
        adv.setStagePath('public_website/advertising')

        self.failUnless(hasattr(public, 'marketing'))
        self.failUnless(hasattr(public, 'advertising'))

        self._populateFolder(mkt.public)
        self._populateFolder(adv.public)

        self.failIf(self._checkStructure(mkt.public))
        self.failIf(self._checkStructure(adv.public))
        self.failUnless(self._checkStructure(public.marketing))
        self.failUnless(self._checkStructure(public.advertising))

        # Check for relative stage URL
        rcu = self.wt.getRelativeStageURL
        self.assertEquals(rcu(mkt.public), '')
        self.assertEquals(rcu(mkt.public.doc1), 'doc1')
        self.assertEquals(rcu(mkt.public.folder1), 'folder1')
        self.assertEquals(rcu(mkt.public.folder2.folder22.folder221),
                          'folder2/folder22/folder221')
        mkt.publishStage()
        adv.publishStage()

        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.advertising)
        self.failIf(errors, errors)

    def testAfterInitialPublishing(self):
        # Add some content after initially publishing a stage
        self.testStagePublishing()
        # Now we have some Stages already setup and under version
        # control.  Now lets add some folders and documents in
        # the staging area and try to publish it.
        mkt_folder11 = self.home.mkt.public.folder1.folder11
        mkt_folder11.invokeFactory('Folder', id='folder111')
        folder111 = mkt_folder11.folder111
        folder111.invokeFactory('Document', id='index_html')
        self.home.mkt.publishStage()
        public = self.public
        pf11 = public.marketing.folder1.folder11
        self.failUnless(hasattr(pf11, 'folder111'))
        self.failUnless(hasattr(pf11.folder111, 'index_html'))

    def testLabeling(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        public = self.public
        public.invokeFactory('Folder', id='marketing')
        mkt.setStagePath('public_website/marketing')
        self._populateFolder(mkt.public)

        # Publish stage, with a label
        errors = self._checkStructure(public.marketing)
        self.failUnless(errors, errors)
        mkt.publishStage(label='test1')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)

        info = mkt.getLastAnnotation()
        self.assertEquals(info['operation'], 'publish+tag')
        self.assertEquals(info['args'].get('label'), mkt.UID()+'-'+'test1')
        self.assertEquals(info['user'],'portal_owner')
        self.assertEquals(info['count'], self.P_COUNT)

        # Check that we don't have any structure
        # inside folder1
        self.failUnless(self._checkStructure(public.marketing.folder1))
        self.failUnless(self._checkStructure(mkt.public.folder1))
        #self.failUnless(self.vt.isCheckedOut(mkt.public.folder1))
        # the default state for objects after publishing has been inverted
        self.failIf(self.vt.isCheckedOut(mkt.public.folder1))
        
        # Populate folder one, and publish with another label
        self._populateFolder(mkt.public.folder1)
        mkt.publishStage(label='test2')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.folder1)
        self.failIf(errors, errors)

        info = mkt.getLastAnnotation()
        self.assertEquals(info['operation'], 'publish+tag')
        self.assertEquals(info['args'].get('label'), mkt.UID()+'-'+'test2')
        self.assertEquals(info['user'], 'portal_owner')
        self.assertEquals(info['count'], self.P_COUNT * 2)

        # Update to previous label, and make sure
        # the structure inside folder1 is gone
        mkt.updateStage(label=mkt.UID()+'-'+'test1')
        errors = self._checkStructure(mkt.public)
        self.failIf(errors, errors)
        errors = self._checkStructure(mkt.public.folder1)
        self.failUnless(errors, errors)

        info = mkt.getLastAnnotation()
        self.assertEquals(info['operation'], 'update')
        self.assertEquals(info['args'].get('label'), mkt.UID()+'-'+'test1')
        self.assertEquals(info['user'], 'portal_owner')
        self.assertEquals(info['count'], self.P_COUNT)

        # Now, update to the other label again and check that
        # the structure inside folder1 is back
        mkt.updateStage(label=mkt.UID()+'-'+'test2')
        errors = self._checkStructure(mkt.public)
        self.failIf(errors, errors)
        errors = self._checkStructure(mkt.public.folder1)
        self.failIf(errors, errors)

        info = mkt.getLastAnnotation()
        self.failIf(info['operation'] != 'update')
        self.failIf(info['args'].get('label') != mkt.UID()+'-'+'test2')
        self.failIf(info['user'] != 'portal_owner')
        self.failIf(info['count'] != self.P_COUNT * 2, info['count'])

    def testRevertToPreviousLabel(self):
        """ Lets make a label or two.  Delete the entire
        stage contents and revert to a previous label.
        """
        self.setupStagingAreas()
        mkt = self.home.mkt
        public = self.public

        public.invokeFactory('Folder', id='marketing')
        mkt.setStagePath('public_website/marketing')
        self._populateFolder(mkt.public)

        # Publish stage, with a label
        self.failUnless(self._checkStructure(public.marketing))
        mkt.publishStage(label='test1')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)

        # Check that path version info has been saved
        # for that label
        stage = self.wt.getStageOf(mkt.public)
        p_info = self.wt.getTagAnnotationForPath(path='',
                                                 label=mkt.UID()+'-'+'test1',
                                                 stage=stage)
        self.assertEquals(len(p_info), 4)

        # Now lets remove everyting and revert to label
        mkt.public.manage_delObjects(mkt.public.objectIds())
        mkt.updateStage(mkt.UID()+'-'+'test1')
        errors = self._checkStructure(mkt.public)
        self.failIf(errors, errors)

    def testPublishDeletePublish(self):
        """ We will create a hierarchy, publish,
        delete a subtree and publish again.
        """
        self.setupStagingAreas()
        mkt = self.home.mkt
        public = self.public

        public.invokeFactory('Folder', id='marketing')
        mkt.setStagePath('public_website/marketing')
        self._populateFolder(mkt.public)
        self._populateFolder(mkt.public.folder1.folder11)

        # Publish stage, with a label
        self.failUnless(self._checkStructure(public.marketing))
        mkt.publishStage(label='test1')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.folder1.folder11)
        self.failIf(errors, errors)

        # Now, delete a subtree and publish again.
        # Content inside 'folder1' of public side
        # should be gone.
        f11 = mkt.public.folder1.folder11
        f11.manage_delObjects(ids=f11.objectIds())
        mkt.publishStage(label='test2')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.folder1.folder11)
        self.failUnless(errors, errors)

        # Revert to label 1, and publish.
        mkt.updateStage(mkt.UID()+'-'+'test1')
        mkt.publishStage()
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.folder1.folder11)
        self.failIf(errors, errors)

        # Create a document at the root of public stage,
        # and another inside folder1.folder11.folder2
        public.marketing.invokeFactory('Document', id='doc3')
        f11_2 = public.marketing.folder1.folder11.folder2
        f11_2.invokeFactory('Document', id='doc4')

        # Now, delete the contents of folder11 in source stage
        # and publish.
        f11 = mkt.public.folder1.folder11
        f11.manage_delObjects(ids=f11.objectIds())
        mkt.publishStage(label='test3')
        errors = self._checkStructure(public.marketing)
        self.failIf(errors, errors)
        errors = self._checkStructure(public.marketing.folder1.folder11)
        self.failUnless(errors, errors)

        # And make sure doc3, doc4 and folder1.folder11.folder2
        # were not deleted.
        p_mkt = public.marketing
        rtv = p_mkt.restrictedTraverse
        self.failUnless(rtv('doc3', None))
        self.failUnless(rtv('folder1/folder11/folder2', None))
        self.failUnless(rtv('folder1/folder11/folder2/doc4', None))

    def testNonVersionable(self):
        # let's check to see if EnSimpleStaging will bomb on
        # non-versionable objects
        self.setupStagingAreas()
        mkt = self.home.mkt
        self.public.invokeFactory('Folder', id='marketing')
        mkt.setStagePath('public_website/marketing')
        # let's create a regular resource
        mkt.public.invokeFactory('Document', id='vers')
        # and a non-versionable one
        mkt.public.invokeFactory('Document', id='nonvers')
        mkt.public.nonvers.__non_versionable__ = True
        # now let's publish the stage and see if our
        # nonvers object hasn't been published
        mkt.publishStage(label='test')
        dest_mkt = self.portal.public_website.marketing
        self.failUnless(dest_mkt.restrictedTraverse('vers', None),
                        "Versionable resource wasn't published")
        #self.failIf(dest_mkt.restrictedTraverse('nonvers', None),
        #                "Non-versionable resource was published")

    def testStagingReferenceability(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        adv = self.home.adv
        sales = self.home.sales
        mkt.public.invokeFactory('InterStageReference', id='aRef')
        # compare received UIDs with known UIDs
        refs = [data[0] for data in
                mkt.public.aRef._referencedStagesData()]
        refs.sort()
        expected = [ob.UID() for ob in mkt, adv, sales]
        expected.sort()
        from pprint import pformat
        self.assertEquals(refs, expected,
                          "available references don't match: %s" %
                          pformat((refs, expected)))

    def testReferencesUnpublisheability(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        self.public.invokeFactory('Folder', id='foo')
        mkt.setStagePath('public_website/foo')
        # let's create some references
        mkt.public.invokeFactory('InterStageReference', id='aRef')
        mkt.public.invokeFactory('Folder', id='aFolder')
        mkt.public.aFolder.invokeFactory('InterStageReference',
                                         id='anotherRef')
        # now let's publish and see if the references were included
        mkt.publishStage(label='test')
        dst = self.portal.public_website.foo
        self.failIf(dst.restrictedTraverse('aRef', None),
                    "reference at public node root was published")
        self.failIf(dst.aFolder.restrictedTraverse('anotherRef', None),
                    "reference deep in the staging area was published")

    def testNestingStagingTargets(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        adv = self.home.adv
        sales = self.home.sales
        mkt.setStagePath('public_website')
        mkt.public.invokeFactory('Folder',id='folder1')
        mkt.publishStage('test1')
        # now we'll land adv content inside the published
        # area of mkt
        adv.setStagePath('public_website/folder1')
        adv.public.invokeFactory('Folder',id='folder11')
        # now we publish adv to see if everything went alright
        adv.publishStage('test2')
        dst_folder1 = self.public.folder1
        obIds = dst_folder1.objectIds()
        self.failIf('index_html' in obIds,
                    "stage reference found inside dest. folder1")
        self.failIf('folder11' not in obIds,
                    'folder11 not found inside dest. folder1')
        dst_folder11 = dst_folder1.folder11
        # and now we change mkt and publish to see if adv stuff survive
        mkt.public.invokeFactory('Folder',id='folder2')
        mkt.publishStage('test3')

        self.failUnless(self.public.restrictedTraverse('folder2',
                                                       None),
                        "folder2 wasn't published")
        dst_folder1 = self.public.folder1
        obIds = dst_folder1.objectIds()
        self.failIf('index_html' in obIds,
                    "stage reference found inside dest. folder1")
        self.failIf('folder11' not in obIds,
                    'folder11 not found inside dest. folder1')
        dst_folder11 = dst_folder1.folder11
        # and now we change adv once more and publish to see if there
        # really were no problems
        adv.public.folder11.invokeFactory('Folder', id='folder111')
        adv.public.invokeFactory('Folder', id='folder12')
        adv.publishStage('test4')

        dst_folder1 = self.public.folder1
        obIds = dst_folder1.objectIds()
        self.failIf('index_html' in obIds,
                    "stage reference found inside dest. folder1")
        self.failIf('folder11' not in obIds,
                    'folder11 not found inside dest. folder1')
        self.failIf('folder12' not in obIds,
                    'folder12 not found inside dest. folder1')
        dst_folder111 = dst_folder1.folder11.restrictedTraverse('folder111',
                                                                None)
        self.failUnless(dst_folder111, "folder111 wasn't published")
        dst_folder11 = dst_folder1.folder11
        # finally, let's try reverting mkt to a previous label
        mkt.updateStage(mkt._globalLabel('test1'))
        mkt.publishStage('test1back')
        # let's check we really went back first
        self.failIf(self.public.restrictedTraverse('folder2',
                                                   None),
                    "folder2 is still published")
        # finally let's see if adv stuff survived
        dst_folder1 = self.public.folder1
        obIds = dst_folder1.objectIds()
        self.failIf('index_html' in obIds,
                    "stage reference found inside dest. folder1")
        self.failIf('folder11' not in obIds,
                    'folder11 not found inside dest. folder1')
        self.failIf('folder12' not in obIds,
                    'folder12 not found inside dest. folder1')
        dst_folder111 = dst_folder1.folder11.restrictedTraverse('folder111',
                                                                None)
        self.failUnless(dst_folder111, "folder111 wasn't published")
        dst_folder11 = dst_folder1.folder11

    def testStagingToolStageListHandling(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        adv = self.home.adv
        sales = self.home.sales
        dst_mkt = makeContent(self.portal, 'Folder', id='mkt')
        dst_sales = makeContent(dst_mkt, 'Folder', id='sales')
        dst_adv = makeContent(dst_sales, 'Folder', id='adv')
        # let's set the stages a little bit out of order
        mkt.setStagePath('mkt')
        adv.setStagePath('mkt/sales/adv')
        sales.setStagePath('mkt/sales')

        for folder in dst_mkt, dst_adv, dst_sales:
            stage = self.wt.getStageFor(folder)
            self.assertEquals(folder, stage, "folder: %s != stage: %s" %
                              (folder.absolute_url(), stage.absolute_url()))

    def testValidStageUIDs(self):
        # this test is here because someone forgot to call the base
        # classes of archetypes while overriding some methods.
        self.setupStagingAreas()
        self.assertNotEquals(self.home.mkt.UID(), self.home.adv.UID())
        self.assertNotEquals(self.home.mkt.UID(), self.home.sales.UID())

    def testTools(self):
        from Products.EnSimpleStaging.tools import locking, versioning, workspaces
        toolIds2Classes = [('portal_workspaces', workspaces.WorkspacesTool),
                           ('portal_versions', versioning.VersioningTool),
                           ('portal_lock', locking.LockingTool)]
        for toolId, toolClass in toolIds2Classes:
            instance = getToolByName(self.portal, toolId)
            self.assertEqual(instance.__class__, toolClass)


    def testNoDuplicateStages(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        adv = self.home.adv
        mkt.setStagePath('public_website')
        self.assertRaises(ValueError, adv.setStagePath, 'public_website')

    def testCatalogIsUpdating(self):
        self.setupStagingAreas()
        cat = self.portal.portal_catalog
        ut = self.portal.plone_utils
        pubsite = self.portal.public_website
        pubsitepath = "/".join(pubsite.getPhysicalPath())
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        f1 = makeContent(mkt.public, 'Folder', id='folder1')
        ut.contentEdit(f1, title='title 1')
        self.assertEquals(f1.Title(), 'title 1')
        # the catalog shouldn't show folder1 inside public_website yet
        self.failIf(cat(id='folder1', path=pubsitepath))
        # now lets publish and see it's there
        mkt.publishStage('test1')
        self.assertEquals(pubsite.folder1.Title(), 'title 1')
        results = cat(id='folder1', path=pubsitepath)
        self.assertEquals(len(results), 1)
        brain = results[0]
        self.assertEquals(brain.Title, 'title 1')
        # now let's put an object inside the published folder
        # to keep it from being erased by recursiveWipeStage
        pubsite.folder1.invokeFactory('Document', id='doc1')
        # and change the source object title
        ut.contentEdit(f1, title='title 2')
        # now lets publish and see if the catalog updated accordingly
        mkt.publishStage('test2')
        self.assertEquals(pubsite.folder1.Title(), 'title 2')
        results = cat(id='folder1', path=pubsitepath)
        self.assertEquals(len(results), 1)
        brain = results[0]
        self.assertEquals(brain.Title, 'title 2')
        # so, now we have a 'title 2' object in the workspace, right?
        nodepath = "/".join(mkt.public.getPhysicalPath())
        results = cat(id='folder1', path=nodepath)
        self.assertEquals(len(results), 1)
        brain = results[0]
        self.assertEquals(brain.Title, 'title 2')
        # in this case, if we revert the stage...
        mkt.updateStage(mkt._globalLabel('test1'))
        # the folder will once again have the old title
        self.assertEquals(f1.Title(), 'title 1')
        # so the catalog info should have the old title too
        nodepath = "/".join(mkt.public.getPhysicalPath())
        results = cat(id='folder1', path=nodepath)
        self.assertEquals(len(results), 1)
        brain = results[0]
        self.assertEquals(brain.Title, 'title 1')

    def testPersistentAttributePublishing(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        pubsite = self.portal.public_website
        mkt.setStagePath('public_website')
        from OFS.Folder import Folder as OFSFolder
        aFolder = makeContent(mkt.public, 'Folder', 'aFolder')
        aFolder.attrFolder = OFSFolder('attrFolder')
        aFolder.attrFolder.manage_afterAdd(aFolder.attrFolder, aFolder)
        aFolder.attrFolder.manage_addProduct['OFSP'].manage_addFolder('a')
        self.subcommit()
        mkt.publishStage('test1')
        self.failUnless('a' in pubsite.aFolder.attrFolder.objectIds())
        aFolder.attrFolder.manage_renameObject('a', 'b')
        # add object to destination to keep it from being wiped
        # during publication
        pubsite.aFolder.invokeFactory('Folder', 'aSubFolder')
        mkt.publishStage('test2')
        self.failIf('a' in pubsite.aFolder.attrFolder.objectIds())
        self.failUnless('b' in pubsite.aFolder.attrFolder.objectIds())

class TestATReferenceStaging(TestStaging):

    def afterSetUp(self):
        TestStaging.afterSetUp(self)
        self.setupStagingAreas()
        self.setupNestedSite()
        tt = self.portal.portal_types
        tt.manage_addTypeInformation('Factory-based Type Information',
                typeinfo_name='Archetypes: SimpleType')

    def testATReferencePublishing(self):
        mkt = self.home.mkt

        # Create the destination folder
        self.public.invokeFactory('Folder', id='foo')

        # Set the stage to point to the newly created folder
        mkt.setStagePath('public_website/foo')

        # Create two pieces of content in there
        mkt.public.invokeFactory(type_name='SimpleType', id='obj1')
        mkt.public.invokeFactory(type_name='SimpleType', id='obj3')
        obj1 = mkt.public.obj1
        obj3 = mkt.public.obj3

        # Create a piece of content *outside* the stages and also
        # outside the nested site.
        self.portal.invokeFactory(type_name='SimpleType', id='obj2')
        obj2 = self.portal.obj2

        # Add some references.
        obj1.addReference(obj2)
        obj1.addReference(obj3)
        self.assertEquals(obj2.getBRefs(), [obj1])
        self.assertEquals(obj1.getRefs(), [obj2, obj3])

        # Publish the stage.
        mkt.publishStage(label='test')
        dst = self.portal.public_website.foo
        orig_uid = obj1.UID()
        new_uid = dst.obj1.UID()
        # Objects in the nested site should keep the same UID.
        self.assertEquals(orig_uid, new_uid)

        new_obj1 = dst.obj1
        new_obj3 = dst.obj3
        obj2 = self.portal.obj2

        # obj2 should still have a reference to obj1
        self.assertEquals(obj2.getBRefs(), [obj1])

        # If we wrap obj2 in the nested site, it should find the
        # staged obj1 instead.
        w_obj2 = obj2.aq_base.__of__(self.portal.public_website)
        self.assertEquals(w_obj2.getBRefs(), [new_obj1])

        self.assertEquals(obj1.getRefs(), [obj2, obj3])
        # Should get None for obj2 as it's outside the
        # nested site, so can't be found by uid_catalog.
        self.assertEquals(new_obj1.getRefs(), [None, new_obj3])

        # Now, let's republish the stage and test it all over again.
        mkt.publishStage(label='test_again')
        new_obj1 = dst.obj1
        new_obj3 = dst.obj3

        # obj2 should still have a reference to obj1
        self.assertEquals(obj2.getBRefs(), [obj1])

        # If we wrap obj2 in the nested site, it should find the
        # staged obj1 instead.
        w_obj2 = obj2.aq_base.__of__(self.portal.public_website)
        self.assertEquals(w_obj2.getBRefs(), [new_obj1])

        self.assertEquals(obj1.getRefs(), [obj2, obj3])
        # Should get None for obj2 as it's outside the
        # nested site, so can't be found by uid_catalog.
        self.assertEquals(new_obj1.getRefs(), [None, new_obj3])

    def testATReferencePublishingSameWorkspace(self):
        mkt = self.home.mkt
        self.public.invokeFactory('Folder', id='foo')
        mkt.setStagePath('public_website/foo')
        mkt.public.invokeFactory(type_name='SimpleType', id='obj1')
        obj1 = mkt.public.obj1
        mkt.public.invokeFactory(type_name='SimpleType', id='obj2')
        obj2 = mkt.public.obj2
        obj1.addReference(obj2)
        self.assertEquals(obj2.getBRefs(), [obj1])
        self.assertEquals(obj1.getRefs(), [obj2])
        mkt.publishStage(label='test')
        dst = self.portal.public_website.foo
        obj1 = dst.obj1
        obj2 = dst.obj2
        self.assertEquals(obj2.getBRefs(), [obj1])
        self.assertEquals(obj1.getRefs(), [obj2])

    def testInterstageReferences(self):
        mkt = self.home.mkt
        adv = self.home.adv
        public = self.public
        public.invokeFactory('Folder', id='mkt')
        public.invokeFactory('Folder', id='adv')
        mkt.setStagePath('public_website/mkt')
        adv.setStagePath('public_website/adv')
        obj1 = makeContent(mkt.public, 'SimpleType', id='obj1')
        obj2 = makeContent(adv.public, 'SimpleType', id='obj2')
        # let's make a reference
        obj1.addReference(obj2)
        # now let's publish the stages
        mkt.publishStage('test1')
        adv.publishStage('test1')
        # so this means that the target obj1 will have a reference to
        # obj2
        new_obj1 = public.mkt.obj1
        new_obj2 = public.adv.obj2
        self.assertEquals(new_obj1.getRefs(), [new_obj2])
        self.assertEquals(new_obj2.getBRefs(), [new_obj1])
        # ok, so if we now republish the object that has the
        # reference, everything should be ok
        mkt.publishStage('test2')
        new_obj1 = public.mkt.obj1
        new_obj2 = public.adv.obj2
        self.assertEquals(new_obj1.getRefs(), [new_obj2])
        self.assertEquals(new_obj2.getBRefs(), [new_obj1])
        # just like when we republish the target of the reference
        adv.publishStage('test2')
        new_obj1 = public.mkt.obj1
        new_obj2 = public.adv.obj2
        self.assertEquals(new_obj1.getRefs(), [new_obj2])
        self.assertEquals(new_obj2.getBRefs(), [new_obj1])

    def testUIDProtection(self):
        self.home.mkt.setStagePath('public_website/mkt')
        obj = makeContent(self.home.mkt.public, 'SimpleType', id='obj')
        obj.addReference(obj)

        from Products.EnSimpleStaging.tools.workspaces import _protectUIDsUnder, _unprotectUIDsUnder
        # let's enable protection
        _protectUIDsUnder(self.home.mkt.public)
        self.failUnless(self.wt._isProtectedUID(obj.UID()),
                        "UID wasn't protected")
        from Products.Archetypes.exceptions import ReferenceException
        rc = self.portal.reference_catalog
        robj = rc.getReferences(obj)[0]
        from Products.EnSimpleStaging.patch import delHook
        self.assertEquals(robj.delHook.im_func, delHook)
        self.assertRaises(ReferenceException, robj.delHook,
                          rc, None, obj)
        self.assertRaises(ReferenceException, robj.delHook,
                          rc, obj, None)
        _unprotectUIDsUnder(self.home.mkt.public)
        self.failIf(self.wt._isProtectedUID(obj.UID()),
                    "UID was still protected")


class TestLocking(TestStaging):

    def testCannotPublishIfLocked(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        lt = self.portal.portal_lock
        wt = self.portal.portal_workspaces
        lt.lock(wt)
        from Products.EnSimpleStaging import StagingError
        self.assertRaises(StagingError, mkt.publishStage, 'test1')

class TestPublicNode(PloneTestCase.PloneTestCase):

    def afterSetUp(self):
        self.loginAsPortalOwner()
        self.qi = self.portal.portal_quickinstaller
        self.qi.installProduct('EnSimpleStaging')
        self.mt = self.portal.portal_membership
        self.wt = self.portal.portal_workspaces
        self.vt = self.portal.portal_versions
        self.acl_users = self.portal.acl_users
        self.home = self.mt.getHomeFolder()
        self.public = makeContent(self.portal, 'Folder', id='public_website')

    def setupStagingAreas(self):
        for area in 'mkt adv sales'.split():
            areaTitle = area + ' staging area'
            self.home.invokeFactory('StagingArea', area, title=areaTitle)

    def testPublicNodeOrderSupport(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        pubnode = mkt.public
        pubsite = self.portal.public_website
        it = getToolByName(self.portal, 'portal_interface')
        iord_str = 'Products.CMFPlone.interfaces.OrderedContainer.IOrderedContainer'
        self.failUnless(it.objectImplements(pubnode, iord_str),
                        'public node not orderable')
        # create subfolders, deploy, check order
        pubnode.invokeFactory('Folder', id='a')
        pubnode.invokeFactory('Folder', id='b')
        self.failUnlessEqual(pubnode.objectIds(), ['a', 'b'])
        mkt.publishStage('test1')
        self.failUnlessEqual(pubsite.objectIds(), ['a', 'b'])
        # move subfolders around, deploy, check order
        pubnode.moveObjectsUp('b')
        self.failUnlessEqual(pubnode.objectIds(), ['b', 'a'])
        mkt.publishStage('test2')
        self.failUnlessEqual(pubsite.objectIds(), ['b', 'a'])

class TestPartialDeployment(TestStaging):

    def _checkLabelsIntegrity(self, workspace):
        labelset = Set()
        for annotation in workspace.getAnnotations():
            label = annotation['label']
            if label in labelset:
                self.fail('label %s was present more than once' % label)
            labelset.add(label)

    def test_getPath2HistoryIdMap(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        pubnode = mkt.public
        # let's add some content and publish
        pubnode.invokeFactory('Folder', id='f1')
        pubnode.f1.invokeFactory('Document', id='d1')
        pubnode.invokeFactory('Folder', id='f2')
        mkt.publishStage('test1')
        stage = self.wt.getStageOf(mkt.public)
        pathmap = self.wt.getPath2HistoryIdMap(stage,
                                               mkt._globalLabel('test1'))
        rightmap = [(('f1',), pubnode.f1.__vc_info__.history_id),
                    (('f1','d1'), pubnode.f1.d1.__vc_info__.history_id),
                    (('f2',), pubnode.f2.__vc_info__.history_id),]
        self.assertEquals(pathmap, rightmap)

    def test_interpretOpcodes(self):
        seq1 = 'a1 aa1 ab1 b1 ba1 bb1'.split()
        seq2 = 'a1 ab1 b1 ba2 bb2 c'.split()
        from Products.EnSimpleStaging.content.stage import interpretOpcodes
        from difflib import SequenceMatcher
        opcodes = SequenceMatcher(None, seq1, seq2).get_opcodes()
        removes, adds = interpretOpcodes(opcodes, seq1, seq2)
        self.assertEquals(removes, 'aa1 ba1 bb1'.split())
        self.assertEquals(adds, 'ba2 bb2 c'.split())

    def test_partial_deployment(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        pubnode = mkt.public
        # let's add some content and publish
        pubnode.invokeFactory('Folder', id='f1')
        pubnode.invokeFactory('Folder', id='f2')
        pubnode.invokeFactory('Document', id='d1')
        pubnode.invokeFactory('Document', id='d4')
        mkt.publishStage('test1')
        # let's sleep a little
        import time
        time.sleep(1)
        pubnode.f2.edit(title='folder 2')
        pubnode.invokeFactory('Document', id='d2')
        pubnode.update(title='foo') # just to make sure it's not deployed
        d1relPath = self.wt.getRelativeStageURL(pubnode.d1) # for later
        self.subcommit()
        pubnode.manage_renameObject('d1', 'd3')
        pubnode.manage_delObjects(ids=['d4'])
        pubnode.manage_clone(pubnode.d3, 'd4')
        # the above should cause some detectable modifications, right?
        lastChanged = [mkt.public.restrictedTraverse(path)
                       for path in mkt.getLastChangedPaths()]
        self.failUnless(pubnode.modified() >
                        mkt.getLastAnnotation()['date'],
                        "pub. node modification wasn't registered: (%s, %s)" %
                        (pubnode.modified(),
                         mkt.getLastAnnotation()['date']))
        self.failIf(pubnode in lastChanged,
                    "pub. node was set to be published")
        self.failUnless(pubnode.f2.modified() >
                        mkt.getLastAnnotation()['date'],
                        "modification wasn't registered: (%s, %s)" %
                        (pubnode.f2.modified(),
                         mkt.getLastAnnotation()['date']))
        self.failUnless(pubnode.f2 in lastChanged,
                        "modification wasn't noticed")
        self.failUnless(pubnode.d2 in lastChanged,
                        "addition wasn't noticed")
        d3relPath = self.wt.getRelativeStageURL(pubnode.d3)
        d4relPath = self.wt.getRelativeStageURL(pubnode.d4)
        mkt.saveStage('save1') # temporary save to let getLastMovements() work
        remPaths, addPaths = mkt.getLastMovements()
        # d1 was moved to d3
        self.failUnless(d1relPath in remPaths)
        self.failUnless(d3relPath in addPaths)
        # d4 was removed and d3 was copied in it's place
        self.failUnless(d4relPath in remPaths)
        self.failUnless(d4relPath in addPaths)
        # let's pretend the last tagging didn't happen
        mkt._si_length.change(-1)
        # that being settled, we need to see if it's publishing the stuff
        # but let's get the suggested next label first
        nl1 = int(mkt.getNextLabel())
        # now we publish
        mkt.publishLastChanged('test2')
        # and check that labels are behaving correctly
        nl2 = int(mkt.getNextLabel())
        self.assertEquals(nl2, nl1 + 1)
        self._checkLabelsIntegrity(mkt)
        # let's check the deployed objects
        self.assertEquals(self.public.f2.Title(), 'folder 2')
        self.failIf('d1' in self.public.objectIds())
        self.failUnless('d3' in self.public.objectIds())
        # and see that the reported number of affected objects is correct
        annot = mkt.getLastAnnotation()
        affected = annot['count']
        self.assertNotEquals(mkt.count(), affected)
        self.assertEquals(affected, 2)
        # and that this last operation can be reverted to:
        self.failUnless('tag' in annot['operation'])

    def _checkVersions(self, obj, numVersions, numLabels=None):
        if numLabels is None:
            numLabels = numVersions
        # check that there are exactly numVersion versions
        # in the version history for an object
        vInfo = self.repo.getVersionInfo(obj)
        histId = vInfo.history_id
        vh = self.repo.getVersionHistory(histId)
        vids = vh.getVersionIds()
        #labels = vh.getLabels()
        self.assertEquals(len(vids), numVersions)

    def testPartialPublishingIsEconomic(self):
        # we've noticed that partial publishing is having some not so
        # nice side effects:
        # * every non-published object is being saved in the
        # VersionRepository again
        #
        # * every published object is being saved in the repository twice
        #
        # basically, all objects are being saved one time more than they
        # should
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        mkt.public.invokeFactory('Folder', id='f1')
        mkt.public.invokeFactory('Folder', id='f2')
        # full-publish the stage a first time
        mkt.publishStage('0')
        # sanity checks:
        # there should be only version of each document in the
        # repository
        self._checkVersions(mkt.public.f1, 1)
        self._checkVersions(mkt.public.f2, 1)
        self.subcommit()
        # let's sleep a little...
        import time
        time.sleep(1)
        # ...change some stuff
        mkt.public.f2.edit(title="Folder number 2")
        self.assertEquals(list(mkt.getLastChangedPaths()),
                          ["/".join(mkt.public.f2.getPhysicalPath())])
        # ...and do a partial publishing
        mkt.publishLastChanged('1')
        # now there should be 1 version and 2 labels of f1...
        self._checkVersions(mkt.public.f1, 1, 2)
        # and 2 versions and 2 labels of f2...
        self._checkVersions(mkt.public.f2, 2)

    def test_worksWithCheckedOutObjects(self):
        # old versions of ESS left objects in the checked-out state
        # partialPublishing must work with them nevertheless
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        mkt.public.invokeFactory('Folder', id='f1')
        mkt.public.invokeFactory('Folder', id='f2')
        # full-publish the stage a first time
        mkt.publishStage('0')
        self.subcommit()
        # let's sleep a little...
        import time
        time.sleep(1)
        # ...leave f1 checked out like old ESS would
        self.repo.checkoutResource(mkt.public.f1)
        mkt.public.f2.edit(title="Folder number 2")
        self.assertEquals(list(mkt.getLastChangedPaths()),
                          ["/".join(mkt.public.f2.getPhysicalPath())])
        # ...and do a partial publishing
        mkt.publishLastChanged('1')
        # check the final state is checked-in
        self.failIf(self.vt.isCheckedOut(mkt.public.f1),
                    "f1 is checked out, shoudl be checked in")

from Products.EnSimpleStaging.Extensions.debugging import cleanUpESSLabels, _removeVersion

class TestCleanUp(TestStaging):

    def getReversibleAnnotations(self, stage):
        return [annotation for annotation in stage.getAnnotations()
                if stage.isReversibleAnnotation(annotation)]

    def testOnly3TagsLeft(self):
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        mkt.public.invokeFactory('Document', id='d1')
        mkt.public.invokeFactory('Document', id='d2')
        # publish the stage a first time
        mkt.publishStage('0')
        # lets gather some data about this publication
        annot0 = mkt.getLastAnnotation()
        label0 = annot0['args']['label']
        doc1VInfo = self.repo.getVersionInfo(mkt.public.d1)
        doc1histId = doc1VInfo.history_id
        doc1vers = doc1VInfo.version_id
        doc1vhistory = self.repo.getVersionHistory(doc1histId)
        doc2VInfo = self.repo.getVersionInfo(mkt.public.d2)
        doc2histId = doc2VInfo.history_id
        # let's check this annotation is reversible
        self.failUnless(mkt.isReversibleAnnotation(annot0))
        # now we delete d2
        mkt.public._delObject('d2')
        mkt.publishStage('1')
        # check that the hist. id for d1 is the same
        doc1VInfo1 = self.repo.getVersionInfo(mkt.public.d1)
        self.assertEquals(doc1histId, doc1VInfo1.history_id)
        # but not the version id
        self.assertNotEquals(doc1vers, doc1VInfo1.version_id)
        # now let's publish some more to make sure deployment 0 falls
        # out of the labels-to-keep range
        mkt.publishStage('2')
        mkt.publishStage('3')
        self.assertEquals(len(self.getReversibleAnnotations(mkt)), 4)
        self.failUnless(mkt.isReversibleAnnotation(annot0))
        cleanUpESSLabels(self.portal, StringIO())
        # let's check that there are less reversible annotations
        self.assertEquals(len(self.getReversibleAnnotations(mkt)), 3)
        # and that the first annotation is no longer reversible
        self.failIf(mkt.isReversibleAnnotation(mkt.getAnnotations()[0]),
                    "1st annotation is still reversible: %s" % mkt.getAnnotations()[0])
        # test if version histories in the repository still have this label
        self.failIf(label0 in doc1vhistory.getLabels())
##        # also check that the version was removed
##        doc1vhistory = self.repo.getVersionHistory(doc1histId)
##        self.failIf(doc1vers in doc1vhistory.getVersionIds())
        # check that the old version doesn't carry any data
        self.assertEquals(doc1vhistory.getVersionById(doc1vers)._data, None,
                          "data hasn't been removed from version")
        # no trace of the old d2 version history should've been left behind
        self.assertRaises(KeyError, self.repo.getVersionHistory, doc2histId)

    def x_test_removeVersion(self):
        d = makeContent(self.home, 'Document', 'doc')
        f = makeContent(self.home, 'Folder', 'f')
        self.repo.applyVersionControl(d)
        vInfo = self.repo.getVersionInfo(d)
        hist_id = vInfo.history_id
        vid0 = vInfo.version_id
        # check in a few times
        self.repo.checkoutResource(d)
        self.repo.checkinResource(d)
        vid1 = self.repo.getVersionInfo(d).version_id
        self.repo.checkoutResource(d)
        self.repo.checkinResource(d)
        vid2 = self.repo.getVersionInfo(d).version_id
        self.repo.checkoutResource(d)
        self.repo.checkinResource(d)
        vid3 = self.repo.getVersionInfo(d).version_id
        vids = (vid0, vid1, vid2, vid3)
        self.assertEquals(len(Set(vids)), 4,
                          "some version_ids are repeated %s" % (vids,))
        # now some sanity checks
        vhistory = self.repo.getVersionHistory(hist_id)
        vers = [vhistory.getVersionById(vid) for vid in vids]
        self.assertEquals(vers[0].prev, None)
        self.assertEquals(vers[3].next, ())
        for i, ver in enumerate(vers[:-1]):
            vid = vids[i]
            nextvid = vids[i+1]
            nextver = vers[i+1]
            # .prev should point to the previous version
            self.assertEquals(nextver.prev, vid,
                              "broken .prev link for version_id %s: %s" %
                              (nextvid, nextver.prev))
            # since we didn't have any branches, the .next tuple
            # should contain only the next version
            self.assertEquals(ver.next, (nextvid,),
                              "broken .next link for version_id %s: %s" %
                              (vid, ver.next))
        # now let's start chopping off versions
        # a guy in the middle
        _removeVersion(vid1, vhistory)
        self.assertEquals(vers[0].next, (vid2,))
        self.assertEquals(vers[2].prev, vid0)
        # a guy in the beginning
        _removeVersion(vid0, vhistory)
        self.assertEquals(vers[2].prev, None)
        # a guy in the end
        _removeVersion(vid3, vhistory)
        self.assertEquals(vers[2].next, ())
        # check that all that is left is version 2
        remaining = list(vhistory._versions.items())
        self.assertEquals(remaining, [ (vid2, vers[2].aq_base) ])

    def testUnchangedVersionStays(self):
        # test if versions that haven't changed are left after the cleanup
        # even if included in the cutoff labels
        self.setupStagingAreas()
        mkt = self.home.mkt
        mkt.setStagePath('public_website')
        mkt.public.invokeFactory('Document', id='d1')
        mkt.public.invokeFactory('Folder', id='f2')
        # publish the stage a first time
        mkt.publishStage('0')
        label0 = mkt.getLastAnnotation()['args']['label']
        doc1VInfo = self.repo.getVersionInfo(mkt.public.d1)
        doc1histId = doc1VInfo.history_id
        doc1vers = doc1VInfo.version_id
        # now let's do some partial publishings
        self.subcommit()
        import time
        time.sleep(1)
        mkt.public.f2.edit(title="title 1")
        mkt.publishLastChanged('1')
        label1 = mkt.getLastAnnotation()['args']['label']
        
        self.subcommit()
        import time
        time.sleep(1)
        mkt.public.f2.edit(title="title 2")
        mkt.publishLastChanged('2')
        label2 = mkt.getLastAnnotation()['args']['label']

        self.subcommit()
        import time
        time.sleep(1)
        mkt.public.f2.edit(title="title 3")
        mkt.publishLastChanged('3')
        label3 = mkt.getLastAnnotation()['args']['label']
        # now we cleanup old deployments.
        cleanUpESSLabels(self.portal, StringIO())
        # and check things worked out
        doc1VInfo = self.repo.getVersionInfo(mkt.public.d1)
        doc1histId = doc1VInfo.history_id
        doc1vers = doc1VInfo.version_id
        # this shouldn't fail
        doc1vhistory = self.repo.getVersionHistory(doc1histId)
        # check the stored info matches
        relpath = self.wt.getRelativeStageURL(mkt.public.d1)
        pathInfo = self.wt._stage_info[mkt._getSourceStageName()][label3].traverse(relpath)
        self.assertEquals((doc1histId, doc1vers),
                          (pathInfo.info.history_id, pathInfo.info.version_id))

class DocTest(TestStaging):
    pass

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    for testclass in (
        TestZopeVersionControl,
        TestEnSimpleStaging,
        TestATReferenceStaging,
        TestLocking,
        TestPublicNode,
        TestPartialDeployment,
        TestCleanUp,
        ):
        suite.addTest(makeSuite(testclass))
    from Testing.ZopeTestCase import FunctionalDocFileSuite
    for doctest in (
        'labels.rst',
        ):
        suite.addTest(FunctionalDocFileSuite(doctest,
                                             test_class=DocTest,
                                             globs=globals()))
    return suite

if __name__ == '__main__':
    framework(descriptions=1, verbosity=1)
